#pragma once

#include <string>
#include <vector>
#include <memory>

#include "tf_data_types.h"

namespace Trueface
{
    class SDK {
        public:
            /**
             * Initialize the SDK using default configuration options.
             *
             * @sa ConfigurationOptions
             */
            SDK();

            /**
             * Initialize the SDK using custom configuration options.
             *
             * @param options custom configuration options.
             */
            SDK(const ConfigurationOptions& options);

            ~SDK();

            /**
             *  Validates the given license token. Need to call this method 
             *  before being able to use the SDK.
             *
             * @param token the license token (if you do not have this talk to
             *        support@trueface.ai).
             * @return Whether the given license token is valid.
             * 
             * @sa isLicensed()
            */
            bool setLicense(std::string token);
#if defined(TARGET_OS_IPHONE) || defined(TARGET_IPHONE_SIMULATOR)
            bool setLicense(std::string token, std::string package_id);
#endif
            /** 
             * Checks whether the given license token is valid and you can use
             * the SDK.
             *
             * @return Whether the given license token is valid.
             * 
             * @sa setLicense()
            */
            bool isLicensed();

            /** 
             * Gets the version-build number of the SDK.
             *
             *
             * @return Version Number as a std::string.
            */
            static std::string getVersion();

            /**
             *  Set the image that is processed by the other methods.
             *
             * @param image an 8-bit image array.
             * @param width the image width.
             * @param height the image height.
             * @param color the image color model, see ColorCode.
             * 
             * @return error code, see ErrorCode.
             * Note, it is highly encouraged to check the return value from setImage before proceeding.
             * If the license is invalid, the INVALID_LICENSE error will be returned.
             *
             * @sa setImage(std::string imageFile)
            */
            ErrorCode setImage(uint8_t* image, uint16_t width, uint16_t height,
                               ColorCode color);

            /** 
             * Set the image that is processed by the other methods.
             *
             * @param imageFile the path of a JPEG or PNG file.
             * 
             * @return error code, see ErrorCode.
             * Note, it is highly encouraged to check the return value from setImage before proceeding.
             * If the license is invalid, the INVALID_LICENSE error will be returned.
             * 
             * @sa setImage(uint8_t* image, uint16_t width, uint16_t height,
             *  ColorCode color)
            */
            ErrorCode setImage(std::string imageFile);

            /** 
             * Detect people and objects in the image.
             *
             * @param[out] boundingBoxes a vector of BoundingBox objects that
             *             gets populated with the detected objects.
             * 
             * @sa getObjectLabelString(ObjectLabel label)
            */
            void detectObjects(std::vector<BoundingBox>& boundingBoxes);

            /** 
             * Estimate body pose.
             *
             * @param[in] boundingBoxes bounding boxes returned by detectObjects
             * @param[out] landmarks a vector of Landmark objects that
             *             represent body joints.
            */
            void estimatePose(const std::vector<BoundingBox>& boundingBoxes, std::vector<std::vector<Landmark>>& landmarks);

            /** 
             * Draw body pose.
             *
             * @param[in] filepath relative or absolute file path without a file extension.
             * @param[in] landmarks a vector of Landmark objects that
             *             represent body joints.
            */
            void drawPose(std::string filepath, const std::vector<std::vector<Landmark>>& landmarks);

            /**
             *  Converts an object label to a std::string. This can be used to 
             *  display the object labels.
             * 
             * @param[in] label the label of a BoundingBox.
             * @return The object label as a std::string.
            */
            std::string getObjectLabelString(ObjectLabel label);

            /**
             * Detect the largest face in the image. See detectFaces() for the detection scale range.
             *
             * @param[out] faceBoxAndLandmarks the bounding box and landmarks 
            *              structure.
             * @param[out] found whether a face was found in the image.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode detectLargestFace(FaceBoxAndLandmarks& faceBoxAndLandmarks,
                                   bool &found);

            /**
             * Detect all the faces in the image. This method has a small false positive rate. To reduce the false
             * positive rate to near zero, filter out faces with score lower than 0.90.
             * 
             * The face detector has a detection scale range of about 5 octaves. \ref ConfigurationOptions.smallestFaceHeight
             * determines the lower of the detection scale range. E.g., setting \ref ConfigurationOptions.smallestFaceHeight to
             * 40 pixels yields the detection scale range of ~40 pixels to 1280 (=40x2^5) pixels.
             *
             * @param[out] faceBoxAndLandmarks a vector of bounding box and landmarks.
             *
             * @return error code, see ErrorCode.
             *              structure.
             */
            ErrorCode detectFaces(std::vector<FaceBoxAndLandmarks>& faceBoxAndLandmarks);

            /**
             * Obtain the 106 face landmarks.
             *
             * @param[in] faceBoxAndLandmarks FaceBoxAndLandmarks returned by detectFaces() or detectLargestFace().
             * @param[out] points a vector of 106 face landmark points.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode getFaceLandmarks(const FaceBoxAndLandmarks& faceBoxAndLandmarks, std::vector<Point<int>>& points);

            /**
             * Align the the detected face to be optimized for passing to 
             *  feature extraction.
             *
             * @param[in] faceBoxAndLandmarks the FaceBoxAndLandmarks returned
             *            by detectLargestFace().
             * @param[out] faceImage the pointer to a uint8_t buffer of 
             *             112x112x3 (=37632) bytes. The aligned face image is 
             *             stored in this buffer.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode extractAlignedFace(const FaceBoxAndLandmarks& faceBoxAndLandmarks,
                                         uint8_t* faceImage,
                                         int marginLeft=0,
                                         int marginTop=0,
                                         int marginRight=0,
                                         int marginBottom=0,
                                         float scale=1.0);

            /**
             * Store the face image in JPEG file.
             *
             * @param[in] faceImage the extracted image by extractAlignedFace().
             * @param[in] filepath relative or absolute file path without a file extension.
            */
            void saveFaceImage(uint8_t* faceImage, std::string filepath);

            /** 
             * Detect the largest face in the image and return its feature
             * vector.
             *
             * @param[out] faceprint a Faceprint object which will contain
             *             the face feature vector.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode getLargestFaceFeatureVector(Faceprint &faceprint);

            /**
             * Extract the face feature vector from an aligned face image.
             *
             * @param[in] alignedFaceImage buffer returned by extractAlignedFace().
             * @param[out] faceprint a Faceprint object which will contain
             *             the face feature vector.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode getFaceFeatureVector(uint8_t* alignedFaceImage, Faceprint& faceprint);

            /**
             * Extract the face feature vectors from the aligned face images.
             * This batch processing method increases throughput on platforms such as cuda.
             *
             * @param[in] alignedFaceImages vector of aligned face image buffers.
             * @param[out] faceprints vector of Faceprint object which will contain
             *             the face feature vectors.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode getFaceFeatureVectors(std::vector<uint8_t*>& alignedFaceImages, std::vector<Faceprint>& faceprints);

            /**
             * Extract the face feature vector from the face box.
             *
             * @param[in] faceBoxAndLandmarks face box returned by detectFaces() or detectLargestFace().
             * @param[out] featureVector vector of floats to contain the
             *             features.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode getFaceFeatureVector(const FaceBoxAndLandmarks& faceBoxAndLandmarks, Faceprint& faceprint);

            /**
             * Compute the similarity between two feature vectors, or how
             * similar two faces are.
             *
             * @param[in] faceprint1 Faceprint of first template to be compared.
             * @param[in] faceprint2 Faceprint of second template to be compared.
             * @param[out] matchProbability the probability the two face feature vectors are a match.
             * @param[out] similarityMeasure the computed similarity measure.
             *
             * @return error code, see ErrorCode.
             * 
             * @sa getLargestFaceFeatureVector
            */
            ErrorCode getSimilarity(const Faceprint& faceprint1,
                                    const Faceprint& faceprint2,
                                    float &matchProbability,
                                    float &similarityMeasure);

            /**
             * Estimate the head pose.
             *
             * @param[in] faceBoxAndLandmarks FaceBoxAndLandmarks returned by detectFaces() or detectLargestFace().
             * @param[out] yaw the rotation angle around the image's vertical axis, in radians.
             * @param[out] pitch the rotation angle around the image's transverse axis, in radians.
             * @param[out] roll the rotation angle around the image's longitudinal axis, in radians.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode estimateHeadOrientation(const FaceBoxAndLandmarks& faceBoxAndLandmarks, float& yaw, float& pitch, float&roll);

            /**
             * Estimate the quality of the face image for recognition.
             *
             * @param[in] alignedFaceImage The array returned by extractAlignedFace().
             * @param[out] quality a value between 0 to 1, 1 being prefect quality for recognition.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode estimateFaceImageQuality(uint8_t* alignedFaceImage, float& quality);

            /**
             * Estimate score for eye blink.
             *
             * @param[out] leftEyeWinkScore score indicating the left eye is open (0 is closed, 1 is open).
             * @param[out] rightEyeWinkScore score indicating the right eye is open (0 is closed, 1 is open).
             *
             * Optimal score threshold for distinguishing between open and closed is 0.4 in normal use and 0.1
             * in low light environments.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode detectBlink(float &leftEyeWinkScore, float &rightEyeWinkScore);

            /**
             * Detect if there is a presentation attack attempt.
             *
             * @param[out] result the predicted results for the face image.
             * @param[out] spoofScore the probability that the image is a spoof attempt
             * (1 indicates real image, 0 indicates spoof attempt).
             * @param[out] threshold the spoof score threshold above which it is considered a spoof attempt.
             *
             * @return error code, see ErrorCode.
            */
            ErrorCode detectSpoof(SpoofLabel& result, float& spoofScore, float threshold = 0.6);

            /**
             *
             * Create a connection to a new or existing database.
             * If the database does not exist, a new one will be created with the provided name.
             * If the NONE DatabaseManagementSystem (memory only) configuration option is selected, this function does
             * not need to be called (and is a harmless no-op).
             *
             * @param databaseConnectionString
             * If SQLITE DatabaseManagementSystem is selected, this should be the filepath to the database.
             * ex. "/myPath/myDatabase.db"
             *
             * @return error code, see ErrorCode.
             */

            ErrorCode createDatabaseConnection(const std::string& databaseConnectionString);

            /**
             *
             * Create a new collection, or load data from an existing collection if one with the
             * provided name already exists in the database.
             *
             * @param[in] collectionName the name of the collection.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode createLoadCollection(const std::string& collectionName);

            /**
             * Enroll a template for a new or existing identity in the collection.
             *
             * @param[in] faceprint the template to enroll in the collection.
             * @param[in] identity the identity corresponding to the template.
             * @param[out] UUID universally unique identifier corresponding to the template.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode enrollTemplate(const Faceprint& faceprint, const std::string& identity, std::string& UUID);

            /**
             * Remove a template from the collection using the UUID.
             *
             * @param[in] UUID the universally unique identifier corresponding to the template
             * to be removed from the collection.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode removeByUUID(const std::string& UUID);

            /**
             * Remove all templates in the collection corresponding to the identity.
             *
             * @param[in] identity the identity to remove from the collection.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode removeByIdentity(const std::string& identity);

            /**
             * Get the top candidate identity in the collection and the corresponding similarity score and match probability.
             *
             * @param[in] faceprint the template to be identified.
             * @param[out] candidate the top candidate identity.
             * @param[out] found set to true if a match is found.
             * @param[in] threshold the similarity score threshold above which it is considered a match.
             * Higher thresholds may result in faster queries.
             * Refer to https://performance.trueface.ai/ when selecting a threshold.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode identifyTopCandidate(const Faceprint& faceprint, Candidate& candidate, bool& found, const float threshold = 0.3f);

            /**
             * Get the top candidate identity in the collection and the corresponding similarity score and match probability
             * for each probe faceprint. Like identifyTopCandidate, but runs search queries in parallel and improves throughput.
             *
             * @param[in] faceprints a vector of templates to be identified.
             * @param[out] candidates the top candidate identity for each probe faceprint.
             * @param[out] found set to true if a match is found for the probe faceprint.
             * @param[in] threshold the similarity score threshold above which it is considered a match.
             * Higher thresholds may result in faster queries.
             * Refer to https://performance.trueface.ai/ when selecting a threshold.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode batchIdentifyTopCandidate(const std::vector<Faceprint>& faceprints, std::vector<Candidate>& candidates,
                    std::vector<bool>& found, const float threshold = 0.3f);

            /**
             * Get a list of the top n candidate identities in the collection and their corresponding similarity scores and match probabilities.
             *
             * @param[in] faceprint the template to be identified.
             * @param[out] candidates a list of the top n matches (or fewer), ordered by descending similarity score.
             * @param[out] found set to true if at leaset one match is found.
             * @param[in] numCandidates max number of candidate identities to return.
             * @param[in] threshold the similarity score threshold above which it is considered a match.
             * Higher thresholds may result in faster queries.
             * Refer to https://performance.trueface.ai/ when selecting a threshold.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode identifyTopCandidates(const Faceprint& faceprint, std::vector<Candidate>& candidates, bool& found,
                            const unsigned int numCandidates, const float threshold = 3.f);

            /**
             * Detect whether the face in the image is wearing a mask or not
             *
             * @param[in] faceBoxAndLandmarks FaceBoxAndLandmarks returned by detectFaces() or detectLargestFace().
             * @param[out] result The predicted result for face image.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode detectMask(const FaceBoxAndLandmarks& faceBoxAndLandmarks, MaskLabel& result);

            /**
             * Detect whether the face in the image is wearing any type of eye glasses or not
             *
             * @param[in] faceBoxAndLandmarks FaceBoxAndLandmarks returned by detectFaces() or detectLargestFace().
             * @param[out] result The predicted result for face image.
             *
             * @return error code, see ErrorCode.
             */
            ErrorCode detectGlasses(const FaceBoxAndLandmarks& faceBoxAndLandmarks, GlassesLabel& result);

    private:
            struct Impl;
            std::unique_ptr<Impl> pImpl;
    };
} // namespace truface
